use std::sync::Arc;
use crate::domain::entity::{Pokemon, PokemonName, PokemonNumber, PokemonTypes};
use crate::repo::pokemon::{InsertError, Repository};

pub struct Request {
    pub(crate) number: u16,
    pub(crate) name: String,
    pub(crate) types: Vec<String>,
}

pub enum Error {
    BadRequest,
    Conflict,
    Unknown,
}

pub struct Response {
    pub number: u16,
    pub name: String,
    pub types: Vec<String>,
}

pub fn execute(repo: Arc<dyn Repository>, req: Request) -> Result<Response, Error> {
    println!("execute");
    match (
        PokemonNumber::try_from(req.number),
        PokemonName::try_from(req.name),
        PokemonTypes::try_from(req.types)
        ) {
        (Ok(number), Ok(name), Ok(types)) => match repo.insert(number, name, types){
            Ok(Pokemon {
                number,
                name,
                types,
               }) => Ok(Response{
                number: u16::from(number),
                name: String::from(name),
                types: Vec::<String>::from(types),
            }),
            Err(InsertError::Conflict) => Err(Error::Conflict),
            Err(InsertError::Error) => Err(Error::Unknown),
        },
        _ => Err(Error::BadRequest),
    }
}

#[cfg(test)]
mod tests {
    use crate::repo::pokemon::InMemoryRepository;
    use super::*;

    #[test]
    fn it_should_return_an_unknown_error_when_an_unexpected_error_happens(){
        let mut repo = Arc::new(InMemoryRepository::new());
        let number = 25;
        let req = Request {
            number: number,
            name: "Pikachu".to_string(),
            types: vec!["Electric".to_string(), ],
        };
        let res = execute(repo, req);
        match res {
            Err(Error::Unknown) => {},
            _ => unreachable!(),
        }
    }

    #[test]
    fn it_should_return_the_pokemon_number_otherwise() {
        let repo = Arc::new(InMemoryRepository::new());
        let num = 25;
        let req = Request {
            number: num,
            name: "Pikachu".to_string(),
            types: vec!["Electric".to_string(), ],
        };
        let res = execute(repo, req);
        match res {
            Ok(Response {
                number,
                name,
                types,
               }) => assert_eq!(number, num),
            _ => unreachable!(),
        }
    }

    #[test]
    fn it_should_return_a_bad_request_error_when_request_is_invalid() {
        let mut repo = Arc::new(InMemoryRepository::new());
        let num = 25;
        let req = Request {
            number: num,
            name: "".to_string(),
            types: vec!["Electric".to_string(), ],
        };
        let res = execute(repo, req);
        match res {
            Ok(Response{
                number,
                name,
                types,
               }) => assert_eq!(number, num),
            Err(Error::BadRequest) => {},
            _ => unreachable!()
        }
    }

    #[test]
    fn it_should_return_a_conflict_error_when_pokemon_number_already_exists() {
        let number = PokemonNumber::try_from(25).unwrap();
        let name = PokemonName::try_from("Pikachu".to_string()).unwrap();
        let types = PokemonTypes::try_from(vec!["Electric".to_string(),]).unwrap();
        let mut repo = Arc::new(InMemoryRepository::new());
        repo.insert(number, name, types);
        let req = Request {
            number: 25,
            name: "Charmander".to_string(),
            types: vec!["File".to_string()],
        };
        let res = execute(repo, req);
        match res {
            Err(Error::Conflict) => {},
            _ => unreachable!(),
        }
    }
}